# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrnetwork

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# XBR Network members.
class Account(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Account()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsAccount(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Account
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Globally unique and static member ID.
    # Account
    def Oid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Account
    def OidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Account
    def OidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Account
    def OidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Timestamp (epoch time in ns) of initial creation of this record.
    # Account
    def Created(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # XBR Network username (must be globally unique on https://xbr.network)
    # Account
    def Username(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # User (primary) email address.
    # Account
    def Email(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Timestamp (epoch time in ns) when the user email was (last) verified or 0 if unverified.
    # Account
    def EmailVerified(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Type of (primary) user crypto wallet in use.
    # Account
    def WalletType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Public address of user crypto wallet in use.
    # Account
    def WalletAddress(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Account
    def WalletAddressAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Account
    def WalletAddressLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Account
    def WalletAddressIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # Block number (on the blockchain) when the member (originally) registered.
    # Account
    def Registered(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Account
    def RegisteredAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Account
    def RegisteredLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Account
    def RegisteredIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # EULA the member agreed to when joining the market (IPFS Multihash string).
    # Account
    def Eula(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Optional member profile (IPFS Multihash string).
    # Account
    def Profile(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Current member level.
    # Account
    def Level(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0


def AccountStart(builder):
    builder.StartObject(11)


def Start(builder):
    AccountStart(builder)


def AccountAddOid(builder, oid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    AccountAddOid(builder, oid)


def AccountStartOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOidVector(builder, numElems: int) -> int:
    return AccountStartOidVector(builder, numElems)


def AccountAddCreated(builder, created):
    builder.PrependUint64Slot(1, created, 0)


def AddCreated(builder, created):
    AccountAddCreated(builder, created)


def AccountAddUsername(builder, username):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(username), 0)


def AddUsername(builder, username):
    AccountAddUsername(builder, username)


def AccountAddEmail(builder, email):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(email), 0)


def AddEmail(builder, email):
    AccountAddEmail(builder, email)


def AccountAddEmailVerified(builder, emailVerified):
    builder.PrependUint64Slot(4, emailVerified, 0)


def AddEmailVerified(builder, emailVerified):
    AccountAddEmailVerified(builder, emailVerified)


def AccountAddWalletType(builder, walletType):
    builder.PrependUint8Slot(5, walletType, 0)


def AddWalletType(builder, walletType):
    AccountAddWalletType(builder, walletType)


def AccountAddWalletAddress(builder, walletAddress):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(walletAddress), 0)


def AddWalletAddress(builder, walletAddress):
    AccountAddWalletAddress(builder, walletAddress)


def AccountStartWalletAddressVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartWalletAddressVector(builder, numElems: int) -> int:
    return AccountStartWalletAddressVector(builder, numElems)


def AccountAddRegistered(builder, registered):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(registered), 0)


def AddRegistered(builder, registered):
    AccountAddRegistered(builder, registered)


def AccountStartRegisteredVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartRegisteredVector(builder, numElems: int) -> int:
    return AccountStartRegisteredVector(builder, numElems)


def AccountAddEula(builder, eula):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(eula), 0)


def AddEula(builder, eula):
    AccountAddEula(builder, eula)


def AccountAddProfile(builder, profile):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(profile), 0)


def AddProfile(builder, profile):
    AccountAddProfile(builder, profile)


def AccountAddLevel(builder, level):
    builder.PrependUint8Slot(10, level, 0)


def AddLevel(builder, level):
    AccountAddLevel(builder, level)


def AccountEnd(builder):
    return builder.EndObject()


def End(builder):
    return AccountEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint

import cbor2
from zlmdb import MapOidFlatBuffers, flatbuffers, table

from cfxdb.gen.realmstore import Publication as PublicationGen


class _Publication(PublicationGen.Publication):
    """
    Expand methods on the class code generated by flatc.

    FIXME: comes up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsPublication(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _Publication()
        x.Init(buf, n + offset)
        return x

    def ArgsAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def KwargsAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PayloadAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def EncKeyAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Publication(object):
    """
    Persisted publication database object.
    """

    ENC_ALGO_NONE = 0
    ENC_ALGO_CRYPTOBOX = 1
    ENC_ALGO_MQTT = 2
    ENC_ALGO_XBR = 3

    ENC_SER_NONE = 0
    ENC_SER_JSON = 1
    ENC_SER_MSGPACK = 2
    ENC_SER_CBOR = 3
    ENC_SER_UBJSON = 4
    ENC_SER_OPAQUE = 5
    ENC_SER_FLATBUFFERS = 6

    __slots__ = (
        "_from_fbs",
        "_timestamp",
        "_publication",
        "_publisher",
        "_topic",
        "_args",
        "_kwargs",
        "_payload",
        "_acknowledge",
        "_retain",
        "_exclude_me",
        "_exclude",
        "_exclude_authid",
        "_exclude_authrole",
        "_eligible",
        "_eligible_authid",
        "_eligible_authrole",
        "_enc_algo",
        "_enc_key",
        "_enc_serializer",
    )

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        self._timestamp = None
        self._publication = None
        self._publisher = None
        self._topic = None
        self._args = None
        self._kwargs = None
        self._payload = None
        self._acknowledge = None
        self._retain = None
        self._exclude_me = None
        self._exclude = None
        self._exclude_authid = None
        self._exclude_authrole = None
        self._eligible = None
        self._eligible_authid = None
        self._eligible_authrole = None
        self._enc_algo = None
        self._enc_key = None
        self._enc_serializer = None

    def marshal(self):
        obj = {
            "timestamp": self.timestamp,
            "publication": self.publication,
            "publisher": self.publisher,
            "topic": self.topic,
            "args": self.args,
            "kwargs": self.kwargs,
            "payload": self.payload,
            "acknowledge": self.acknowledge,
            "retain": self.retain,
            "exclude_me": self.exclude_me,
            "exclude": self.exclude,
            "exclude_authid": self.exclude_authid,
            "exclude_authrole": self.exclude_authrole,
            "eligible": self.eligible,
            "eligible_authid": self.eligible_authid,
            "eligible_authrole": self.eligible_authrole,
            "enc_algo": self.enc_algo,
            "enc_key": self.enc_key,
            "enc_serializer": self.enc_serializer,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def timestamp(self):
        """
        Timestamp when the publication was accepted by the broker. Epoch time in ns.

        :returns: epoch time in ns
        :rtype: int
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = self._from_fbs.Timestamp()
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        assert value is None or type(value) == int
        self._timestamp = value

    @property
    def publication(self):
        """
        WAMP publication ID that was assigned by the broker.

        :returns: publication ID
        :rtype: int
        """
        if self._publication is None and self._from_fbs:
            self._publication = self._from_fbs.Publication()
        return self._publication

    @publication.setter
    def publication(self, value):
        assert value is None or type(value) == int
        self._publication = value

    @property
    def publisher(self):
        """
        WAMP session ID of the publisher.

        :returns: publisher ID
        :rtype: int
        """
        if self._publisher is None and self._from_fbs:
            self._publisher = self._from_fbs.Publisher()
        return self._publisher

    @publisher.setter
    def publisher(self, value):
        assert value is None or type(value) == int
        self._publisher = value

    @property
    def topic(self):
        """
        The WAMP or application URI of the PubSub topic the event was published to.

        :returns: topic (URI) published to
        :rtype: str
        """
        if self._topic is None and self._from_fbs:
            self._topic = self._from_fbs.Topic().decode("utf8")
        return self._topic

    @topic.setter
    def topic(self, value):
        assert value is None or type(value) == str
        self._topic = value

    #
    # args, kwargs, payload
    #

    @property
    def args(self):
        """
        Positional values for application-defined event payload.

        :returns: positional arguments (app payload) of the event (if any)
        :rtype: None or list
        """
        if self._args is None and self._from_fbs:
            if self._from_fbs.ArgsLength():
                self._args = cbor2.loads(bytes(self._from_fbs.ArgsAsBytes()))
        return self._args

    @args.setter
    def args(self, value):
        assert value is None or type(value) == list
        self._args = value

    @property
    def kwargs(self):
        """
        Keyword values for application-defined event payload.

        :returns: keyword arguments (app payload) of the event (if any)
        :rtype: None or dict
        """
        if self._kwargs is None and self._from_fbs:
            if self._from_fbs.KwargsLength():
                self._kwargs = cbor2.loads(bytes(self._from_fbs.KwargsAsBytes()))
        return self._kwargs

    @kwargs.setter
    def kwargs(self, value):
        assert value is None or type(value) == dict
        self._kwargs = value

    @property
    def payload(self):
        """
        Alternative, transparent payload. If given, ``args`` and ``kwargs`` must be left unset.

        :returns: Transparent binary payload (see ``enc_algo``) if applicable
        :rtype: None or bytes
        """
        if self._payload is None and self._from_fbs:
            if self._from_fbs.PayloadLength():
                self._payload = self._from_fbs.PayloadAsBytes()
        return self._payload

    @payload.setter
    def payload(self, value):
        assert value is None or type(value) == bytes
        self._payload = value

    #
    # acknowledge, retain, exclude_me
    #

    @property
    def acknowledge(self):
        """
        If ``True``, the broker was asked to acknowledge the publication with a success or error response.

        :returns: acknowledge flag
        :rtype: None or bool
        """
        if self._acknowledge is None and self._from_fbs:
            self._acknowledge = self._from_fbs.Acknowledge()
        return self._acknowledge

    @acknowledge.setter
    def acknowledge(self, value):
        assert value is None or type(value) == bool
        self._acknowledge = value

    @property
    def retain(self):
        """
        If ``True``, the broker was requested to retain this event.

        :returns: retain flag
        :rtype: None or bool
        """
        if self._retain is None and self._from_fbs:
            self._retain = self._from_fbs.Retain()
        return self._retain

    @retain.setter
    def retain(self, value):
        assert value is None or type(value) == bool
        self._retain = value

    @property
    def exclude_me(self):
        """
        If ``True``, the broker was asked to exclude the publisher from receiving the event.

        :returns: exclude_me flag
        :rtype: None or bool
        """
        if self._exclude_me is None and self._from_fbs:
            self._exclude_me = self._from_fbs.ExcludeMe()
        return self._exclude_me

    @exclude_me.setter
    def exclude_me(self, value):
        assert value is None or type(value) == bool
        self._exclude_me = value

    #
    # exclude, exclude_authid, exclude_authrole
    #

    @property
    def exclude(self):
        """
        List of WAMP session IDs to exclude from receiving this event.

        :returns: list of excluded session IDs
        :rtype: list[int]
        """
        if self._exclude is None and self._from_fbs:
            if self._from_fbs.ExcludeLength():
                exclude = []
                for j in range(self._from_fbs.ExcludeLength()):
                    exclude.append(self._from_fbs.Exclude(j))
                self._exclude = exclude
        return self._exclude

    @exclude.setter
    def exclude(self, value):
        assert value is None or type(value) == list
        if value:
            for x in value:
                assert type(x) == int
        self._exclude = value

    @property
    def exclude_authid(self):
        """
        List of WAMP authids to exclude from receiving this event.

        :returns: list of excluded authids
        :rtype: list[str]
        """
        if self._exclude_authid is None and self._from_fbs:
            if self._from_fbs.ExcludeAuthidLength():
                exclude_authid = []
                for j in range(self._from_fbs.ExcludeAuthidLength()):
                    exclude_authid.append(self._from_fbs.ExcludeAuthid(j).decode("utf8"))
                self._exclude_authid = exclude_authid
        return self._exclude_authid

    @exclude_authid.setter
    def exclude_authid(self, value):
        assert value is None or type(value) == list
        if value:
            for x in value:
                assert type(x) == str
        self._exclude_authid = value

    @property
    def exclude_authrole(self):
        """
        List of WAMP authroles to exclude from receiving this event.

        :returns: list of excluded authroles
        :rtype: list[str]
        """
        if self._exclude_authrole is None and self._from_fbs:
            if self._from_fbs.ExcludeAuthroleLength():
                exclude_authrole = []
                for j in range(self._from_fbs.ExcludeAuthroleLength()):
                    exclude_authrole.append(self._from_fbs.ExcludeAuthrole(j).decode("utf8"))
                self._exclude_authrole = exclude_authrole
        return self._exclude_authrole

    @exclude_authrole.setter
    def exclude_authrole(self, value):
        assert value is None or type(value) == list
        if value:
            for x in value:
                assert type(x) == str
        self._exclude_authrole = value

    #
    # eligible, eligible_authid, eligible_authrole
    #

    @property
    def eligible(self):
        """
        List of WAMP session IDs eligible to receive this event.

        :returns: list of eligible session IDs
        :rtype: list[int]
        """
        if self._eligible is None and self._from_fbs:
            if self._from_fbs.EligibleLength():
                eligible = []
                for j in range(self._from_fbs.EligibleLength()):
                    eligible.append(self._from_fbs.Eligible(j))
                self._eligible = eligible
        return self._eligible

    @eligible.setter
    def eligible(self, value):
        assert value is None or type(value) == list
        if value:
            for x in value:
                assert type(x) == int
        self._eligible = value

    @property
    def eligible_authid(self):
        """
        List of WAMP authids eligible to receive this event.

        :returns: list of eligible authids
        :rtype: list[str]
        """
        if self._eligible_authid is None and self._from_fbs:
            if self._from_fbs.EligibleAuthidLength():
                eligible_authid = []
                for j in range(self._from_fbs.EligibleAuthidLength()):
                    eligible_authid.append(self._from_fbs.EligibleAuthid(j).decode("utf8"))
                self._eligible_authid = eligible_authid
        return self._eligible_authid

    @eligible_authid.setter
    def eligible_authid(self, value):
        assert value is None or type(value) == list
        if value:
            for x in value:
                assert type(x) == str
        self._eligible_authid = value

    @property
    def eligible_authrole(self):
        """
        List of WAMP authroles eligible to receive this event.

        :returns: list of eligible authroles
        :rtype: list[str]
        """
        if self._eligible_authrole is None and self._from_fbs:
            if self._from_fbs.EligibleAuthroleLength():
                eligible_authrole = []
                for j in range(self._from_fbs.EligibleAuthroleLength()):
                    eligible_authrole.append(self._from_fbs.EligibleAuthrole(j).decode("utf8"))
                self._eligible_authrole = eligible_authrole
        return self._eligible_authrole

    @eligible_authrole.setter
    def eligible_authrole(self, value):
        assert value is None or type(value) == list
        if value:
            for x in value:
                assert type(x) == str
        self._eligible_authrole = value

    #
    # encryption
    #

    @property
    def enc_algo(self):
        """
        When using payload transparency, the encoding algorithm that was used to encode the payload.

        :returns: payload encryption algorithm
        :rtype: int
        """
        if self._enc_algo is None and self._from_fbs:
            self._enc_algo = self._from_fbs.EncAlgo()
        return self._enc_algo

    @enc_algo.setter
    def enc_algo(self, value):
        assert value is None or value in [self.ENC_ALGO_CRYPTOBOX, self.ENC_ALGO_MQTT, self.ENC_ALGO_XBR]
        self._enc_algo = value

    @property
    def enc_key(self):
        """
        When using payload transparency with an encryption algorithm, the payload encryption key.

        :returns: payload key
        :rtype: None or bytes
        """
        if self._enc_key is None and self._from_fbs:
            if self._from_fbs.EncKeyLength():
                self._enc_key = self._from_fbs.EncKeyAsBytes()
        return self._enc_key

    @enc_key.setter
    def enc_key(self, value):
        assert value is None or type(value) == bytes
        self._enc_key = value

    @property
    def enc_serializer(self):
        """
        When using payload transparency, the payload object serializer that was used encoding the payload.

        :returns: payload serializer
        :rtype: int
        """
        if self._enc_serializer is None and self._from_fbs:
            self._enc_serializer = self._from_fbs.EncSerializer()
        return self._enc_serializer

    @enc_serializer.setter
    def enc_serializer(self, value):
        assert value is None or value in [
            self.ENC_SER_JSON,
            self.ENC_SER_MSGPACK,
            self.ENC_SER_CBOR,
            self.ENC_SER_UBJSON,
        ]
        self._enc_serializer = value

    @staticmethod
    def cast(buf):
        return Publication(_Publication.GetRootAsPublication(buf, 0))

    def build(self, builder):
        args = self.args
        if args:
            args = builder.CreateString(cbor2.dumps(args))

        kwargs = self.kwargs
        if kwargs:
            kwargs = builder.CreateString(cbor2.dumps(kwargs))

        payload = self.payload
        if payload:
            payload = builder.CreateString(payload)

        topic = self.topic
        if topic:
            topic = builder.CreateString(topic)

        enc_key = self.enc_key
        if enc_key:
            enc_key = builder.CreateString(enc_key)

        # exclude: [int]
        exclude = self.exclude
        if exclude:
            PublicationGen.PublicationStartExcludeAuthidVector(builder, len(exclude))
            for session in reversed(exclude):
                builder.PrependUint64(session)
            exclude = builder.EndVector()

        # exclude_authid: [string]
        exclude_authid = self.exclude_authid
        if exclude_authid:
            _exclude_authid = []
            for authid in exclude_authid:
                _exclude_authid.append(builder.CreateString(authid))
            PublicationGen.PublicationStartExcludeAuthidVector(builder, len(_exclude_authid))
            for o in reversed(_exclude_authid):
                builder.PrependUOffsetTRelative(o)
            exclude_authid = builder.EndVector()

        # exclude_authrole: [string]
        exclude_authrole = self.exclude_authrole
        if exclude_authid:
            _exclude_authrole = []
            for authrole in exclude_authrole:
                _exclude_authrole.append(builder.CreateString(authrole))
            PublicationGen.PublicationStartExcludeAuthroleVector(builder, len(_exclude_authrole))
            for o in reversed(_exclude_authrole):
                builder.PrependUOffsetTRelative(o)
            exclude_authrole = builder.EndVector()

        # eligible: [int]
        eligible = self.eligible
        if eligible:
            PublicationGen.PublicationStartEligibleAuthidVector(builder, len(eligible))
            for session in reversed(eligible):
                builder.PrependUint64(session)
            eligible = builder.EndVector()

        # eligible_authid: [string]
        eligible_authid = self.eligible_authid
        if eligible_authid:
            _eligible_authid = []
            for authid in eligible_authid:
                _eligible_authid.append(builder.CreateString(authid))
            PublicationGen.PublicationStartEligibleAuthidVector(builder, len(_eligible_authid))
            for o in reversed(_eligible_authid):
                builder.PrependUOffsetTRelative(o)
            eligible_authid = builder.EndVector()

        # eligible_authrole: [string]
        eligible_authrole = self.eligible_authrole
        if eligible_authrole:
            _eligible_authrole = []
            for authrole in eligible_authrole:
                _eligible_authrole.append(builder.CreateString(authrole))
            PublicationGen.PublicationStartEligibleAuthroleVector(builder, len(_eligible_authrole))
            for o in reversed(_eligible_authrole):
                builder.PrependUOffsetTRelative(o)
            eligible_authrole = builder.EndVector()

        # now start and build a new object ..
        PublicationGen.PublicationStart(builder)

        if self.timestamp:
            PublicationGen.PublicationAddTimestamp(builder, self.timestamp)
        if self.publication:
            PublicationGen.PublicationAddPublication(builder, self.publication)
        if self.publisher:
            PublicationGen.PublicationAddPublisher(builder, self.publisher)

        if topic:
            PublicationGen.PublicationAddTopic(builder, topic)

        if args:
            PublicationGen.PublicationAddArgs(builder, args)
        if kwargs:
            PublicationGen.PublicationAddKwargs(builder, kwargs)
        if payload is not None:
            PublicationGen.PublicationAddPayload(builder, payload)

        if self.acknowledge is not None:
            PublicationGen.PublicationAddAcknowledge(builder, self.acknowledge)
        if self.retain is not None:
            PublicationGen.PublicationAddRetain(builder, self.retain)
        if self.exclude_me is not None:
            PublicationGen.PublicationAddExcludeMe(builder, self.exclude_me)

        if exclude:
            PublicationGen.PublicationAddExclude(builder, exclude)
        if exclude_authid:
            PublicationGen.PublicationAddExcludeAuthid(builder, exclude_authid)
        if exclude_authrole:
            PublicationGen.PublicationAddExcludeAuthrole(builder, exclude_authrole)

        if eligible:
            PublicationGen.PublicationAddEligible(builder, eligible)
        if eligible_authid:
            PublicationGen.PublicationAddEligibleAuthid(builder, eligible_authid)
        if eligible_authrole:
            PublicationGen.PublicationAddEligibleAuthrole(builder, eligible_authrole)

        if self.enc_algo:
            PublicationGen.PublicationAddEncAlgo(builder, self.enc_algo)
        if enc_key:
            PublicationGen.PublicationAddEncKey(builder, enc_key)
        if self.enc_serializer:
            PublicationGen.PublicationAddEncSerializer(builder, self.enc_serializer)

        # finish the object.
        final = PublicationGen.PublicationEnd(builder)

        return final


@table("dd04931a-753b-4fde-8140-d66b93519c73", build=Publication.build, cast=Publication.cast)
class Publications(MapOidFlatBuffers):
    """
    Persisted publications archive.

    Map :class:`zlmdb.MapOidFlatBuffers` from ``publication`` to :class:`cfxdb.eventstore.Publication`.
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: mrealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class RouterWorkerGroupClusterPlacement(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RouterWorkerGroupClusterPlacement()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsRouterWorkerGroupClusterPlacement(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # RouterWorkerGroupClusterPlacement
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # OID of this worker-to-cluster placement.
    # RouterWorkerGroupClusterPlacement
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # OID of (router) worker group this placement applies to.
    # RouterWorkerGroupClusterPlacement
    def WorkerGroupOid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # OID of cluster onto which the worker is placed.
    # RouterWorkerGroupClusterPlacement
    def ClusterOid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # OID of the node onto which the worker is placed.
    # RouterWorkerGroupClusterPlacement
    def NodeOid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Local worker name on the node onto which the worker is placed.
    # RouterWorkerGroupClusterPlacement
    def WorkerName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def RouterWorkerGroupClusterPlacementStart(builder):
    builder.StartObject(5)


def Start(builder):
    RouterWorkerGroupClusterPlacementStart(builder)


def RouterWorkerGroupClusterPlacementAddOid(builder, oid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    RouterWorkerGroupClusterPlacementAddOid(builder, oid)


def RouterWorkerGroupClusterPlacementAddWorkerGroupOid(builder, workerGroupOid):
    builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(workerGroupOid), 0)


def AddWorkerGroupOid(builder, workerGroupOid):
    RouterWorkerGroupClusterPlacementAddWorkerGroupOid(builder, workerGroupOid)


def RouterWorkerGroupClusterPlacementAddClusterOid(builder, clusterOid):
    builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(clusterOid), 0)


def AddClusterOid(builder, clusterOid):
    RouterWorkerGroupClusterPlacementAddClusterOid(builder, clusterOid)


def RouterWorkerGroupClusterPlacementAddNodeOid(builder, nodeOid):
    builder.PrependStructSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(nodeOid), 0)


def AddNodeOid(builder, nodeOid):
    RouterWorkerGroupClusterPlacementAddNodeOid(builder, nodeOid)


def RouterWorkerGroupClusterPlacementAddWorkerName(builder, workerName):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(workerName), 0)


def AddWorkerName(builder, workerName):
    RouterWorkerGroupClusterPlacementAddWorkerName(builder, workerName)


def RouterWorkerGroupClusterPlacementEnd(builder):
    return builder.EndObject()


def End(builder):
    return RouterWorkerGroupClusterPlacementEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import json

from .account import Accounts, IndexAccountsByEmail, IndexAccountsByUsername, IndexAccountsByWallet
from .userkey import IndexUserKeyByAccount, UserKeys
from .vaction import VerifiedActions


class Schema(object):
    """
    XBR Network backend database schema.
    """

    accounts: Accounts
    """
    Member accounts database table :class:`xbrnetwork.Accounts`.
    """

    idx_accounts_by_username: IndexAccountsByUsername
    """
    Index "by username" of member accounts :class:`xbrnetwork.IndexAccountsByUsername`.
    """

    idx_accounts_by_email: IndexAccountsByEmail
    """
    Index "by email" of member accounts :class:`xbrnetwork.IndexAccountsByEmail`.
    """

    idx_accounts_by_wallet: IndexAccountsByWallet
    """
    Index "by wallet" of member accounts :class:`xbrnetwork.IndexAccountsByWallet`.
    """

    verified_actions: VerifiedActions
    """
    Verified actions database table :class:`xbrnetwork.VerifiedActions`.
    """

    user_keys: UserKeys
    """
    User client keys database table :class:`xbrnetwork.UserKeys`.
    """

    idx_user_key_by_account: IndexUserKeyByAccount
    """
    Index "by pubkey" of user keys :class:`xbrnetwork.IndexUserKeyByAccount`.
    """

    EXPORTED = None
    """
    """

    def __init__(self, db):
        self.db = db

    def export(self, txn, fd):
        """

        :param txn:
        :param fd:
        :return:
        """
        recs = 0
        db_data = {}
        for table in self.EXPORTED:
            table_data = []
            for key, val in table.select(txn, return_keys=True, return_values=True):
                table_data.append((key, val.marshal() if val else None))
                recs += 1
            db_data[table.__name__] = table_data
        db_data = json.dumps(db_data, ensure_ascii=False)
        fd.write(db_data)
        return recs

    @staticmethod
    def attach(db):
        """
        Attach database schema to database instance.

        :param db: Database to attach schema to.
        :type db: :class:`zlmdb.Database`
        """
        schema = Schema(db)

        schema.accounts = db.attach_table(Accounts)

        schema.idx_accounts_by_username = db.attach_table(IndexAccountsByUsername)
        schema.accounts.attach_index("idx1", schema.idx_accounts_by_username, lambda account: account.username)

        schema.idx_accounts_by_email = db.attach_table(IndexAccountsByEmail)
        schema.accounts.attach_index("idx2", schema.idx_accounts_by_email, lambda account: account.email)

        schema.idx_accounts_by_wallet = db.attach_table(IndexAccountsByWallet)
        schema.accounts.attach_index("idx3", schema.idx_accounts_by_wallet, lambda account: account.wallet_address)

        schema.verified_actions = db.attach_table(VerifiedActions)

        schema.user_keys = db.attach_table(UserKeys)

        schema.idx_user_key_by_account = db.attach_table(IndexUserKeyByAccount)
        schema.user_keys.attach_index(
            "idx1", schema.idx_user_key_by_account, lambda user_key: (user_key.owner, user_key.created)
        )

        schema.EXPORTED = [schema.accounts, schema.verified_actions, schema.user_keys]

        return schema



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from .activation_token import ActivationToken
from .organization import Organization
from .user import User
from .user_mrealm_role import UserMrealmRole


class Schema(object):
    """
    user database schema for ZLMDB.
    """

    def __init__(self, db):
        self.db = db

    activation_tokens: ActivationToken
    """
    CFC user activation tokens.
    """

    organizations: Organization
    """
    CFC organization.
    """

    users: User
    """
    CFC users.
    """

    user_mrealm_roles: UserMrealmRole
    """
    CFC user roles.
    """

    @staticmethod
    def attach(db):
        """
        Factory to create a schema from attaching to a database. The schema tables
        will be automatically mapped as persistant maps and attached to the
        database slots.

        :param db: zlmdb.Database
        :return: object of Schema
        """
        schema = Schema(db)

        schema.activation_tokens = db.attach_table(ActivationToken)

        schema.organizations = db.attach_table(Organization)

        schema.users = db.attach_table(User)

        schema.user_mrealm_roles = db.attach_table(UserMrealmRole)

        return schema



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import Optional
from uuid import UUID


class ApplicationRealmRoleAssociation(object):
    """
    Association of a role with an application realm.
    """

    def __init__(self, arealm_oid: Optional[UUID] = None, role_oid: Optional[UUID] = None, _unknown=None):
        self.arealm_oid = arealm_oid
        self.role_oid = role_oid
        self._unknown = _unknown
        """

        :param arealm_oid: Object ID of the application realm this association applies to.

        :param role_oid: Object ID of the role this association applies to.
        """

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if other.arealm_oid != self.arealm_oid:
            return False
        if other.role_oid != self.role_oid:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = {
            "arealm_oid": str(self.arealm_oid),
            "role_oid": str(self.role_oid),
        }
        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`WebService`
        """
        assert type(data) == dict

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["arealm_oid", "role_oid"]:
                _unknown[k] = data[k]

        arealm_oid = None
        if "arealm_oid" in data:
            assert type(data["arealm_oid"]) == str
            arealm_oid = UUID(data["arealm_oid"])

        role_oid = None
        if "role_oid" in data:
            assert type(data["role_oid"]) == str
            role_oid = UUID(data["role_oid"])

        obj = ApplicationRealmRoleAssociation(arealm_oid=arealm_oid, role_oid=role_oid, _unknown=_unknown)

        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import List, Optional
from uuid import UUID

import numpy as np

from cfxdb.mrealm.cluster import Cluster


class RouterCluster(Cluster):
    """
    A router cluster is able to run (application) realms, and is hosted on a groups of router workers
    kept in sync and meshed via router-to-router links.
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        name: Optional[str] = None,
        status: Optional[int] = None,
        owner_oid: Optional[UUID] = None,
        changed: Optional[np.datetime64] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of this router cluster.

        :param label: Optional user label of this router cluster.

        :param description: Optional user description of this router cluster.

        :param tags: Optional list of user tags on this router cluster.
        """
        Cluster.__init__(
            self,
            oid=oid,
            label=label,
            description=description,
            tags=tags,
            name=name,
            status=status,
            owner_oid=owner_oid,
            changed=changed,
        )

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not Cluster.__eq__(self, other):
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = Cluster.marshal(self)
        obj.update({})
        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`ManagementRealm`
        """
        assert type(data) == dict

        obj = Cluster.parse(data)
        data = obj._unknown

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in []:
                _unknown[k] = data[k]

        obj = RouterCluster(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            name=obj.name,
            status=obj.status,
            owner_oid=obj.owner_oid,
            changed=obj.changed,
            _unknown=_unknown,
        )

        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################
import pprint
import uuid

import numpy as np
from zlmdb import MapBytes20TimestampUuid, MapBytes20Uuid, MapUuidFlatBuffers, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbr import Market as MarketGen


class _MarketGen(MarketGen.Market):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsMarket(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _MarketGen()
        x.Init(buf, n + offset)
        return x

    def MarketAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def CreatedAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def OwnerAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def CoinAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def MakerAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ProviderSecurityAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ConsumerSecurityAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def MarketFeeAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def TidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def SignatureAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Market(object):
    """
    ``XBRNetwork.Market`` database object.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # [uint8] (uuid)
        self._market = None

        # uint64 (timestamp)
        self._timestamp = None

        # uint256
        self._created = None

        # uint32
        self._seq = None

        # [uint8] (address)
        self._owner = None

        # [uint8] (address)
        self._coin = None

        # string (multihash)
        self._terms = None

        # string (multihash)
        self._meta = None

        # [uint8] (address)
        self._maker = None

        # [uint8] (uint256)
        self._provider_security = None

        # [uint8] (uint256)
        self._consumer_security = None

        # [uint8] (uint256)
        self._market_fee = None

        # [uint8] (ethhash)
        self._tid = None

        # [uint8] (ethsig)
        self._signature = None

    def marshal(self) -> dict:
        obj = {
            "market": self.market.bytes if self.market else None,
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "created": pack_uint256(self.created) if self.created else None,
            "seq": self.seq,
            "owner": bytes(self.owner) if self.owner else None,
            "coin": bytes(self.coin) if self.coin else None,
            "terms": self.terms,
            "meta": self.meta,
            "maker": bytes(self.maker) if self.maker else None,
            "provider_security": pack_uint256(self.provider_security) if self.provider_security else None,
            "consumer_security": pack_uint256(self.consumer_security) if self.consumer_security else None,
            "market_fee": pack_uint256(self.market_fee) if self.market_fee else None,
            "tid": bytes(self.tid) if self.tid else None,
            "signature": bytes(self.signature) if self.signature else None,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def market(self) -> uuid.UUID:
        """
        The unique ID of the market.
        """
        if self._market is None and self._from_fbs:
            if self._from_fbs.MarketLength():
                _market = self._from_fbs.MarketAsBytes()
                self._market = uuid.UUID(bytes=bytes(_market))
        return self._market

    @market.setter
    def market(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._market = value

    @property
    def timestamp(self) -> np.datetime64:
        """
        Database transaction time (epoch time in ns) of insert or last update.
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def created(self) -> int:
        """
        Block number (on the blockchain) when the actor (originally) joined the market.
        """
        if self._created is None and self._from_fbs:
            if self._from_fbs.CreatedLength():
                _created = self._from_fbs.CreatedAsBytes()
                self._created = unpack_uint256(bytes(_created))
            else:
                self._created = 0
        return self._created

    @created.setter
    def created(self, value: int):
        assert value is None or type(value) == int
        self._created = value

    @property
    def seq(self) -> int:
        """
        Global market sequence number.
        """
        if self._seq is None and self._from_fbs:
            self._seq = self._from_fbs.Seq()
        return self._seq or 0

    @seq.setter
    def seq(self, value: int):
        assert value is None or type(value) == int
        self._seq = value

    @property
    def owner(self) -> bytes:
        """
        Market owner.
        """
        if self._owner is None and self._from_fbs:
            if self._from_fbs.OwnerLength():
                self._owner = self._from_fbs.OwnerAsBytes()
        return self._owner

    @owner.setter
    def owner(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._owner = value

    @property
    def coin(self) -> bytes:
        """
        Market coin.
        """
        if self._coin is None and self._from_fbs:
            if self._from_fbs.CoinLength():
                self._coin = self._from_fbs.CoinAsBytes()
        return self._coin

    @coin.setter
    def coin(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._coin = value

    @property
    def terms(self) -> str:
        """
        The XBR market terms set by the market owner. IPFS Multihash pointing to a ZIP archive file with market documents.
        """
        if self._terms is None and self._from_fbs:
            terms = self._from_fbs.Terms()
            if terms:
                self._terms = terms.decode("utf8")
        return self._terms

    @terms.setter
    def terms(self, value: str):
        assert value is None or type(value) == str
        self._terms = value

    @property
    def meta(self) -> str:
        """
        The XBR market metadata published by the market owner. IPFS Multihash pointing to a RDF/Turtle file with market metadata.
        """
        if self._meta is None and self._from_fbs:
            meta = self._from_fbs.Meta()
            if meta:
                self._meta = meta.decode("utf8")
        return self._meta

    @meta.setter
    def meta(self, value):
        assert value is None or type(value) == str
        self._meta = value

    @property
    def maker(self) -> bytes:
        """
        The address of the XBR market maker that will run this market. The delegate of the market owner.
        """
        if self._maker is None and self._from_fbs:
            if self._from_fbs.MakerLength():
                self._maker = self._from_fbs.MakerAsBytes()
        return self._maker

    @maker.setter
    def maker(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._maker = value

    @property
    def provider_security(self) -> int:
        """
        The amount of XBR tokens a XBR provider joining the market must deposit.
        """
        if self._provider_security is None and self._from_fbs:
            if self._from_fbs.ProviderSecurityLength():
                _provider_security = self._from_fbs.ProviderSecurityAsBytes()
                self._provider_security = unpack_uint256(bytes(_provider_security))
            else:
                self._provider_security = 0
        return self._provider_security

    @provider_security.setter
    def provider_security(self, value: int):
        assert value is None or type(value) == int
        self._provider_security = value

    @property
    def consumer_security(self) -> int:
        """
        The amount of XBR tokens a XBR consumer joining the market must deposit.
        """
        if self._consumer_security is None and self._from_fbs:
            if self._from_fbs.ConsumerSecurityLength():
                _consumer_security = self._from_fbs.ConsumerSecurityAsBytes()
                self._consumer_security = unpack_uint256(bytes(_consumer_security))
            else:
                self._consumer_security = 0
        return self._consumer_security

    @consumer_security.setter
    def consumer_security(self, value: int):
        assert value is None or type(value) == int
        self._consumer_security = value

    @property
    def market_fee(self) -> int:
        """
        The fee taken by the market (beneficiary is the market owner). The fee is a percentage of the revenue of the XBR Provider that receives XBR Token paid for transactions. The fee must be between 0% (inclusive) and 99% (inclusive), and is expressed as a fraction of the total supply of XBR tokens.
        """
        if self._market_fee is None and self._from_fbs:
            if self._from_fbs.MarketFeeLength():
                _market_fee = self._from_fbs.MarketFeeAsBytes()
                self._market_fee = unpack_uint256(bytes(_market_fee))
            else:
                self._market_fee = 0
        return self._market_fee

    @market_fee.setter
    def market_fee(self, value):
        assert value is None or type(value) == int
        self._market_fee = value

    @property
    def tid(self) -> bytes:
        """
        Transaction hash of the transaction this change was committed to the blockchain under.
        """
        if self._tid is None and self._from_fbs:
            if self._from_fbs.TidLength():
                self._tid = self._from_fbs.TidAsBytes()
        return self._tid

    @tid.setter
    def tid(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 32)
        self._tid = value

    @property
    def signature(self) -> bytes:
        """
        When signed off-chain and submitted via ``XBRMarket.createMarketFor``.
        """
        if self._signature is None and self._from_fbs:
            if self._from_fbs.SignatureLength():
                self._signature = self._from_fbs.SignatureAsBytes()
        return self._signature

    @signature.setter
    def signature(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._signature = value

    @staticmethod
    def cast(buf):
        return Market(_MarketGen.GetRootAsMarket(buf, 0))

    def build(self, builder):
        market = self.market.bytes if self.market else None
        if market:
            market = builder.CreateString(market)

        created = self.created
        if created:
            created = builder.CreateString(pack_uint256(created))

        owner = self.owner
        if owner:
            owner = builder.CreateString(owner)

        coin = self.coin
        if coin:
            coin = builder.CreateString(coin)

        terms = self.terms
        if terms:
            terms = builder.CreateString(terms)

        meta = self.meta
        if meta:
            meta = builder.CreateString(meta)

        maker = self.maker
        if maker:
            maker = builder.CreateString(maker)

        provider_security = self.provider_security
        if provider_security:
            provider_security = builder.CreateString(pack_uint256(provider_security))

        consumer_security = self.consumer_security
        if consumer_security:
            consumer_security = builder.CreateString(pack_uint256(consumer_security))

        market_fee = self.market_fee
        if market_fee:
            market_fee = builder.CreateString(pack_uint256(market_fee))

        tid = self.tid
        if tid:
            tid = builder.CreateString(tid)

        signature = self.signature
        if signature:
            signature = builder.CreateString(signature)

        MarketGen.MarketStart(builder)

        if market:
            MarketGen.MarketAddMarket(builder, market)

        if self.timestamp:
            MarketGen.MarketAddTimestamp(builder, int(self.timestamp))

        if created:
            MarketGen.MarketAddProviderSecurity(builder, created)

        if self.seq:
            MarketGen.MarketAddSeq(builder, self.seq)

        if owner:
            MarketGen.MarketAddOwner(builder, owner)

        if coin:
            MarketGen.MarketAddCoin(builder, coin)

        if terms:
            MarketGen.MarketAddTerms(builder, terms)

        if meta:
            MarketGen.MarketAddMeta(builder, meta)

        if maker:
            MarketGen.MarketAddMaker(builder, maker)

        if provider_security:
            MarketGen.MarketAddProviderSecurity(builder, provider_security)

        if consumer_security:
            MarketGen.MarketAddConsumerSecurity(builder, consumer_security)

        if market_fee:
            MarketGen.MarketAddMarketFee(builder, market_fee)

        if tid:
            MarketGen.MarketAddTid(builder, tid)

        if signature:
            MarketGen.MarketAddSignature(builder, signature)

        final = MarketGen.MarketEnd(builder)

        return final


@table("861b0942-0c3f-4d41-bc35-d8c86af0b2c9", build=Market.build, cast=Market.cast)
class Markets(MapUuidFlatBuffers):
    """
    Markets table, mapping from ``market_id|UUID`` to :class:`cfxdb.xbr.Market`
    """


@table("7c3d67b4-35a3-449f-85a6-2695636fc63e")
class IndexMarketsByOwner(MapBytes20TimestampUuid):
    """
    Markets-by-owner index with ``(owner_adr|bytes[20], created|int) -> market_id|UUID`` mapping.
    """


@table("4f50a97a-4531-4eab-a91b-45cc42b3dd21")
class IndexMarketsByActor(MapBytes20TimestampUuid):
    """
    Markets-by-actor index with ``(actor_adr|bytes[20], joined|int) -> market_id|UUID`` mapping.
    """


@table("d511774c-0c7b-4d3f-a2de-6748c072a56f")
class IndexMarketsByMaker(MapBytes20Uuid):
    """
    Markets-by-maker index with ``maker_adr|bytes[20] -> market_id|UUID`` mapping.
    """



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import zlmdb
from zlmdb import MapStringUuid, MapTimestampUuidFlatBuffers, MapUuidCbor, MapUuidStringUuid, MapUuidUuidCbor, table

from cfxdb.mrealm import ManagementRealm, Node
from cfxdb.usage import MasterNodeUsage
from cfxdb.user import ActivationToken, Organization, User, UserMrealmRole

__all__ = (
    "ManagementRealms",
    "IndexManagementRealmByName",
    "Nodes",
    "IndexNodesByPubkey",
    "IndexNodesByAuthid",
    "Users",
    "IndexUsersByName",
    "IndexUsersByPubkey",
    "IndexUsersByEmail",
    "UserMrealmRoles",
    "ActivationTokens",
    "IndexActivationTokensByAuthidPubkey",
    "Organizations",
    "IndexOrganizationsByName",
    "UsageRecords",
    "GlobalSchema",
)


@table("1219e71c-a62c-415a-bd15-ddf45e3a658b", marshal=ManagementRealm.marshal, parse=ManagementRealm.parse)
class ManagementRealms(MapUuidCbor):
    """
    Management realm database configuration object (``oid -> mrealm``).
    """


@table("1d2e8045-ea2b-4456-be4a-7a234d3622d6")
class IndexManagementRealmByName(MapStringUuid):
    """
    Index of management realms by name (``mrealm_name -> mrealm_oid``).
    """


@table("ae89d956-273a-4ce3-b63d-52b07ae35742", marshal=Node.marshal, parse=Node.parse)
class Nodes(MapUuidCbor):
    """
    Nodes paired with management realms (``node_oid -> node``).
    """


@table("1336c623-5f38-4397-ad5b-2e6b716b57b0")
class IndexNodesByPubkey(MapStringUuid):
    """
    Index of nodes by node public key (``node_pubkey -> node_oid``).
    """


@table("6d071a33-0577-4f72-a2e1-11182f60ab9c")
class IndexNodesByAuthid(MapUuidStringUuid):
    """
    Index of nodes by management realm and node authid (``(mrealm_oid, authid) -> node_oid``).
    """


@table("fa1ed0fc-304e-4f66-8092-d901df1735e4", marshal=User.marshal, parse=User.parse)
class Users(MapUuidCbor):
    """
    Users registered in this domain / master node (``user_oid -> user``).
    """


@table("aa2754e5-a859-4986-8749-1299828dc6e1")
class IndexUsersByName(MapStringUuid):
    """
    Index of users by name (``user_name -> user_oid``).
    """


@table("882a24e4-90cc-4823-94fe-c1d938daffe6")
class IndexUsersByPubkey(MapStringUuid):
    """
    Index of users by pubkey (``user_pubkey -> user_oid``).
    """


@table("933447a3-dd79-4599-bd9a-e0d88d9b84cb")
class IndexUsersByEmail(MapStringUuid):
    """
    Index of users by email address (``user_email -> user_oid``).
    """


@table("eccdfc57-5632-4ad4-9c2b-2ac11e9d389f", marshal=UserMrealmRole.marshal, parse=UserMrealmRole.parse)
class UserMrealmRoles(MapUuidUuidCbor):
    """
    User roles on management realms (``(mrealm_oid, role_oid) -> role``).
    """


@table("c968886e-a2e2-490c-bc2a-6b684c3130f6", marshal=ActivationToken.marshal, parse=ActivationToken.parse)
class ActivationTokens(MapUuidCbor):
    """
    User activations (``activation_oid -> activation``).
    """


@table("0f6a9014-2e39-4cfd-9f2b-f6ffd3d3deca")
class IndexActivationTokensByAuthidPubkey(MapStringUuid):
    """
    Index of activation tokes by user authid and pubkey (``(user_authid, user_pubkey) -> activation_oid``).
    """


@table("ae2fe53f-f8ec-4484-8a8f-cabdf1b38358", marshal=Organization.marshal, parse=Organization.parse)
class Organizations(MapUuidCbor):
    """
    Organizations defined in this domain / master node (``organization_oid -> organization``).
    """


@table("3fb82ab2-430d-43a1-8200-fcd6355d0410")
class IndexOrganizationsByName(MapStringUuid):
    """
    Index of organizations by name (``organization_name -> organization_oid``).
    """


@table("e38f7bf1-2514-400c-8c30-a979b2138503", build=MasterNodeUsage.build, cast=MasterNodeUsage.cast)
class UsageRecords(MapTimestampUuidFlatBuffers):
    """
    Usage metering records (``(timestamp, usage_oid) -> usage_record``).
    """


class GlobalSchema(object):
    """
    Global (master node wide) database schema.
    """

    def __init__(self, db):
        self.db = db

    nodes: Nodes
    """
    Nodes.

    * Database table :class:`cfxdb.globalschema.Nodes`
    """

    idx_nodes_by_pubkey: IndexNodesByPubkey
    """
    Index on nodes (by pubkey).

    * Database table :class:`cfxdb.globalschema.IndexNodesByPubkey`
    """

    idx_nodes_by_authid: IndexNodesByAuthid
    """
    Index on nodes (by authid).

    * Database table :class:`cfxdb.globalschema.IndexNodesByAuthid`
    """

    organizations: Organizations
    """
    Organizations.

    * Database table :class:`cfxdb.globalschema.Organizations`
    """

    idx_organizations_by_name: IndexOrganizationsByName
    """
    Index on organizations (by name).

    * Database table :class:`cfxdb.globalschema.IndexOrganizationsByName`
    """

    users: Users
    """
    Users.

    * Database table :class:`cfxdb.globalschema.Users`
    """

    idx_users_by_pubkey: IndexUsersByPubkey
    """
    Index on users (by pubkey).

    * Database table :class:`cfxdb.globalschema.IndexUsersByPubkey`
    """

    idx_users_by_email: IndexUsersByEmail
    """
    Index on users (by email).

    * Database table :class:`cfxdb.globalschema.IndexUsersByEmail`
    """

    activation_tokens: ActivationTokens
    """
    User activation tokens.

    * Database table :class:`cfxdb.globalschema.ActivationTokens`
    """

    idx_act_tokens_by_authid_pubkey: IndexActivationTokensByAuthidPubkey
    """
    Index on user activation tokens (by authid, pubkey).

    * Database table :class:`cfxdb.globalschema.IndexActivationTokensByAuthidPubkey`
    """

    mrealms: ManagementRealms
    """
    Management realms.

    * Database table :class:`cfxdb.globalschema.ManagementRealms`
    """

    idx_mrealms_by_name: IndexManagementRealmByName
    """
    Index on management realms (by name).

    * Database table :class:`cfxdb.globalschema.IndexManagementRealmByName`
    """

    users_mrealm_roles: UserMrealmRoles
    """
    User roles map (by mrealm_oid, user_oid to UserRoles).

    * Database table :class:`cfxdb.globalschema.UserMrealmRoles`
    """

    usage: UsageRecords
    """
    Usage metering records (by timestamp).

    * Database table :class:`cfxdb.globalschema.UsageRecords`
    """

    @staticmethod
    def attach(db: zlmdb.Database) -> object:
        """
        Factory to create a schema from attaching to a database. The schema tables
        will be automatically mapped as persistant maps and attached to the
        database slots.

        :param db: Database to which to attach this schema.

        :returns: Open database handle with attached schema, an instance of :class:`GlobalSchema`.
        """
        schema = GlobalSchema(db)

        # Organizations
        #
        schema.organizations = db.attach_table(Organizations)

        schema.idx_organizations_by_name = db.attach_table(IndexOrganizationsByName)
        schema.organizations.attach_index("idx1", schema.idx_organizations_by_name, lambda org: org.name)

        # Users
        #
        schema.users = db.attach_table(Users)

        schema.idx_users_by_pubkey = db.attach_table(IndexUsersByPubkey)
        schema.users.attach_index("idx1", schema.idx_users_by_pubkey, lambda user: user.pubkey, nullable=True)

        schema.idx_users_by_email = db.attach_table(IndexUsersByEmail)
        schema.users.attach_index("idx2", schema.idx_users_by_email, lambda user: user.email, nullable=True)

        schema.activation_tokens = db.attach_table(ActivationTokens)

        schema.idx_act_tokens_by_authid_pubkey = db.attach_table(IndexActivationTokensByAuthidPubkey)
        schema.activation_tokens.attach_index(
            "idx1", schema.idx_act_tokens_by_authid_pubkey, lambda token: token.email + token.pubkey
        )

        # Management Realms
        #
        schema.mrealms = db.attach_table(ManagementRealms)

        schema.idx_mrealms_by_name = db.attach_table(IndexManagementRealmByName)
        schema.mrealms.attach_index("idx1", schema.idx_mrealms_by_name, lambda mrealm: mrealm.name)

        schema.users_mrealm_roles = db.attach_table(UserMrealmRoles)

        # Nodes
        #
        schema.nodes = db.attach_table(Nodes)

        schema.idx_nodes_by_pubkey = db.attach_table(IndexNodesByPubkey)
        schema.nodes.attach_index("idx1", schema.idx_nodes_by_pubkey, lambda node: node.pubkey)

        schema.idx_nodes_by_authid = db.attach_table(IndexNodesByAuthid)
        schema.nodes.attach_index(
            "idx2", schema.idx_nodes_by_authid, lambda node: (node.mrealm_oid, node.authid), nullable=True
        )

        # Usage metering
        #
        schema.usage = db.attach_table(UsageRecords)

        return schema



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid

import numpy as np
from zlmdb import MapUuidFlatBuffers, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbrmm import Transaction as TransactionGen


class _TransactionGen(TransactionGen.Transaction):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsTransaction(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _TransactionGen()
        x.Init(buf, n + offset)
        return x

    def TidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def OfferAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def AmountAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PaymentChannelAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PayingChannelAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def KeyAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def BuyerPubkeyAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PaymentChannelAfterAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PayingChannelAfterAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PaymentMmSigAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PaymentDelSigAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PayingMmSigAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PayingDelSigAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Transaction(object):
    """
    Data encryption key buy-sell transactions.
    """

    STATUS_NONE = 0
    """
    Unset
    """

    STATUS_INFLIGHT = 1
    """
    The transaction is currently in-flight
    """

    STATUS_FAILED = 2
    """
    The transaction has completed with error (it failed)
    """

    STATUS_SUCCESS = 3
    """
    The transaction has completed with success
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # uint8[] (uuid)
        self._tid = None

        # uint64
        self._created = None

        # uint32
        self._created_payment_channel_seq = None

        # uint32
        self._created_paying_channel_seq = None

        # uint8[] (uuid)
        self._offer = None

        # uint8[] (uint256)
        self._amount = None

        # uint8[] (uuid)
        self._payment_channel = None

        # uint8[] (uuid)
        self._paying_channel = None

        # uint8
        self._state = None

        # uint64
        self._completed = None

        # uint32
        self._completed_payment_channel_seq = None

        # uint32
        self._completed_paying_channel_seq = None

        # [uint8] (uuid)
        self._key = None

        # [uint8] (uint256)
        self._buyer_pubkey = None

        # [uint8] (uint256)
        self._payment_channel_after = None

        # [uint8] (uint256)
        self._paying_channel_after = None

        # [uint8] (ethsig)
        self._payment_mm_sig = None

        # [uint8] (ethsig)
        self._payment_del_sig = None

        # [uint8] (ethsig)
        self._paying_mm_sig = None

        # [uint8] (ethsig)
        self._paying_del_sig = None

    def marshal(self):
        obj = {
            "tid": str(self.tid) if self.tid else None,
            "created": self.created,
            "created_payment_channel_seq": self._created_payment_channel_seq,
            "created_paying_channel_seq": self._created_paying_channel_seq,
            "offer": str(self.offer) if self.offer else None,
            "amount": pack_uint256(self.amount) if self.amount else 0,
            "payment_channel": self.payment_channel.bytes if self.payment_channel else None,
            "paying_channel": self.paying_channel.bytes if self.paying_channel else None,
            "state": self.state,
            "completed": self.completed,
            "completed_payment_channel_seq": self._completed_payment_channel_seq,
            "completed_paying_channel_seq": self._completed_paying_channel_seq,
            "key": self.key.bytes if self.key else None,
            "buyer_pubkey": self.buyer_pubkey,
            "payment_channel_after": pack_uint256(self.payment_channel_after) if self.payment_channel_after else None,
            "paying_channel_after": pack_uint256(self.paying_channel_after) if self.paying_channel_after else None,
            "payment_mm_sig": self.payment_mm_sig,
            "payment_del_sig": self.payment_del_sig,
            "paying_mm_sig": self.paying_mm_sig,
            "paying_del_sig": self.paying_del_sig,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def tid(self) -> uuid.UUID:
        """
        ID of the transaction.
        """
        if self._tid is None and self._from_fbs:
            if self._from_fbs.OfferLength():
                _tid = self._from_fbs.TidAsBytes()
                self._tid = uuid.UUID(bytes=bytes(_tid))
        return self._tid

    @tid.setter
    def tid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._tid = value

    @property
    def created(self) -> np.datetime64:
        """
        Creation time of the transaction (epoch time in ns).
        """
        if self._created is None and self._from_fbs:
            self._created = np.datetime64(self._from_fbs.Created(), "ns")
        return self._created

    @created.setter
    def created(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._created = value

    @property
    def created_payment_channel_seq(self) -> int:
        """
        Sequence number of the created-state transaction within the payment channel.
        """
        if self._created_payment_channel_seq is None and self._from_fbs:
            self._created_payment_channel_seq = self._from_fbs.CreatedPaymentChannelSeq()
        return self._created_payment_channel_seq or 0

    @created_payment_channel_seq.setter
    def created_payment_channel_seq(self, value: int):
        assert value is None or type(value) == int
        self._created_payment_channel_seq = value

    @property
    def created_paying_channel_seq(self) -> int:
        """
        Sequence number of the created-state transaction within the paying channel.
        """
        if self._created_paying_channel_seq is None and self._from_fbs:
            self._created_paying_channel_seq = self._from_fbs.CreatedPayingChannelSeq()
        return self._created_paying_channel_seq or 0

    @created_paying_channel_seq.setter
    def created_paying_channel_seq(self, value: int):
        assert value is None or type(value) == int
        self._created_paying_channel_seq = value

    @property
    def offer(self) -> uuid.UUID:
        """
        ID of the data encryption key offer this transaction is for.
        """
        if self._offer is None and self._from_fbs:
            if self._from_fbs.OfferLength():
                _offer = self._from_fbs.OfferAsBytes()
                self._offer = uuid.UUID(bytes=bytes(_offer))
        return self._offer

    @offer.setter
    def offer(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._offer = value

    @property
    def amount(self) -> int:
        """
        Transaction amount in XBR.
        """
        if self._amount is None and self._from_fbs:
            if self._from_fbs.AmountLength():
                _amount = self._from_fbs.AmountAsBytes()
                self._amount = unpack_uint256(bytes(_amount))
            else:
                self._amount = 0
        return self._amount

    @amount.setter
    def amount(self, value: int):
        assert value is None or type(value) == int
        self._amount = value

    @property
    def payment_channel(self) -> uuid.UUID:
        """
        Address of the payment channel (of the buyer) this transaction is transacting on.
        """
        if self._payment_channel is None and self._from_fbs:
            if self._from_fbs.PaymentChannelLength():
                self._payment_channel = uuid.UUID(bytes=bytes(self._from_fbs.PaymentChannelAsBytes()))
        return self._payment_channel

    @payment_channel.setter
    def payment_channel(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._payment_channel = value

    @property
    def paying_channel(self) -> uuid.UUID:
        """
        Address of the paying channel (of the seller) this transaction is transacting on.
        """
        if self._payment_channel is None and self._from_fbs:
            if self._from_fbs.PayingChannelLength():
                self._payment_channel = uuid.UUID(bytes=bytes(self._from_fbs.PayingChannelAsBytes()))
        return self._payment_channel

    @paying_channel.setter
    def paying_channel(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._payment_channel = value

    @property
    def state(self) -> int:
        """
        State of the transaction: TransactionState.INFLIGHT when created, and then either TransactionState.SUCCESS or TransactionState.FAILED.
        """
        if self._state is None and self._from_fbs:
            self._state = self._from_fbs.State()
        return self._state or 0

    @state.setter
    def state(self, value: int):
        assert value is None or type(value) == int
        self._state = value

    @property
    def completed(self) -> np.datetime64:
        """
        Completion time of the transaction (epoch time in ns)
        """
        if self._completed is None and self._from_fbs:
            self._completed = np.datetime64(self._from_fbs.Completed(), "ns")
        return self._completed

    @completed.setter
    def completed(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._completed = value

    @property
    def completed_payment_channel_seq(self) -> int:
        """
        Sequence number of the completed-state transaction within the payment channel.
        """
        if self._completed_payment_channel_seq is None and self._from_fbs:
            self._completed_payment_channel_seq = self._from_fbs.CompletedPaymentChannelSeq()
        return self._completed_payment_channel_seq or 0

    @completed_payment_channel_seq.setter
    def completed_payment_channel_seq(self, value: int):
        assert value is None or type(value) == int
        self._completed_payment_channel_seq = value

    @property
    def completed_paying_channel_seq(self) -> int:
        """
        Sequence number of the completed-state transaction within the paying channel.
        """
        if self._completed_paying_channel_seq is None and self._from_fbs:
            self._completed_paying_channel_seq = self._from_fbs.CompletedPayingChannelSeq()
        return self._completed_paying_channel_seq or 0

    @completed_paying_channel_seq.setter
    def completed_paying_channel_seq(self, value: int):
        assert value is None or type(value) == int
        self._completed_paying_channel_seq = value

    @property
    def key(self) -> uuid.UUID:
        """
        ID of the data encryption key sold under the transaction.
        """
        if self._key is None and self._from_fbs:
            if self._from_fbs.KeyLength():
                _key = self._from_fbs.KeyAsBytes()
                self._key = uuid.UUID(bytes=bytes(_key))
        return self._key

    @key.setter
    def key(self, value):
        assert value is None or isinstance(value, uuid.UUID)
        self._key = value

    @property
    def buyer_pubkey(self) -> bytes:
        """
        Buyer public key (Ed25519).
        """
        if self._buyer_pubkey is None and self._from_fbs:
            if self._from_fbs.BuyerPubkeyLength():
                self._buyer_pubkey = self._from_fbs.BuyerPubkeyAsBytes()
        return self._buyer_pubkey

    @buyer_pubkey.setter
    def buyer_pubkey(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 32)
        self._buyer_pubkey = value

    @property
    def payment_channel_after(self) -> int:
        """
        Payment channel balance after transaction.
        """
        if self._payment_channel_after is None and self._from_fbs:
            if self._from_fbs.PaymentChannelAfterLength():
                _payment_channel_after = self._from_fbs.PaymentChannelAfterAsBytes()
                self._payment_channel_after = unpack_uint256(bytes(_payment_channel_after))
            else:
                self._payment_channel_after = 0
        return self._payment_channel_after

    @payment_channel_after.setter
    def payment_channel_after(self, value: int):
        assert value is None or type(value) == int
        self._payment_channel_after = value

    @property
    def paying_channel_after(self) -> int:
        """
        Payment channel balance after transaction.
        """
        if self._paying_channel_after is None and self._from_fbs:
            if self._from_fbs.PayingChannelAfterLength():
                _paying_channel_after = self._from_fbs.PayingChannelAfterAsBytes()
                self._paying_channel_after = unpack_uint256(bytes(_paying_channel_after))
            else:
                self._paying_channel_after = 0
        return self._paying_channel_after

    @paying_channel_after.setter
    def paying_channel_after(self, value: int):
        assert value is None or type(value) == int
        self._paying_channel_after = value

    @property
    def payment_mm_sig(self) -> bytes:
        """
        Payment channel market maker transaction signature.
        """
        if self._payment_mm_sig is None and self._from_fbs:
            if self._from_fbs.PaymentMmSigLength():
                self._payment_mm_sig = self._from_fbs.PaymentMmSigAsBytes()
        return self._payment_mm_sig

    @payment_mm_sig.setter
    def payment_mm_sig(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._payment_mm_sig = value

    @property
    def payment_del_sig(self) -> bytes:
        """
        Payment channel (buyer) delegate transaction signature.
        """
        if self._payment_del_sig is None and self._from_fbs:
            if self._from_fbs.PaymentDelSigLength():
                self._payment_del_sig = self._from_fbs.PaymentDelSigAsBytes()
        return self._payment_del_sig

    @payment_del_sig.setter
    def payment_del_sig(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._payment_del_sig = value

    @property
    def paying_mm_sig(self) -> bytes:
        """
        Paying channel market maker transaction signature.
        """
        if self._paying_mm_sig is None and self._from_fbs:
            if self._from_fbs.PayingMmSigLength():
                self._paying_mm_sig = self._from_fbs.PayingMmSigAsBytes()
        return self._paying_mm_sig

    @paying_mm_sig.setter
    def paying_mm_sig(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._paying_mm_sig = value

    @property
    def paying_del_sig(self) -> bytes:
        """
        Paying channel (seller) delegate transaction signature.
        """
        if self._paying_del_sig is None and self._from_fbs:
            if self._from_fbs.PayingDelSigLength():
                self._paying_del_sig = self._from_fbs.PayingDelSigAsBytes()
        return self._paying_del_sig

    @paying_del_sig.setter
    def paying_del_sig(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._paying_del_sig = value

    @staticmethod
    def cast(buf):
        return Transaction(_TransactionGen.GetRootAsTransaction(buf, 0))

    def build(self, builder):
        tid = self.tid.bytes if self.tid else None
        if tid:
            tid = builder.CreateString(tid)

        offer = self.offer.bytes if self.offer else None
        if offer:
            offer = builder.CreateString(offer)

        amount = self.amount
        if amount:
            amount = builder.CreateString(pack_uint256(amount))

        payment_channel = self.payment_channel.bytes if self.payment_channel else None
        if payment_channel:
            payment_channel = builder.CreateString(payment_channel)

        paying_channel = self.paying_channel.bytes if self.paying_channel else None
        if paying_channel:
            paying_channel = builder.CreateString(paying_channel)

        key = self.key
        if key:
            key = builder.CreateString(key.bytes)

        buyer_pubkey = self.buyer_pubkey
        if buyer_pubkey:
            buyer_pubkey = builder.CreateString(buyer_pubkey)

        payment_channel_after = self.payment_channel_after
        if payment_channel_after:
            payment_channel_after = builder.CreateString(pack_uint256(payment_channel_after))

        paying_channel_after = self.paying_channel_after
        if paying_channel_after:
            paying_channel_after = builder.CreateString(pack_uint256(paying_channel_after))

        payment_mm_sig = self.payment_mm_sig
        if payment_mm_sig:
            payment_mm_sig = builder.CreateString(payment_mm_sig)

        payment_del_sig = self.payment_del_sig
        if payment_del_sig:
            payment_del_sig = builder.CreateString(payment_del_sig)

        paying_mm_sig = self.paying_mm_sig
        if paying_mm_sig:
            paying_mm_sig = builder.CreateString(paying_mm_sig)

        paying_del_sig = self.paying_del_sig
        if paying_del_sig:
            paying_del_sig = builder.CreateString(paying_del_sig)

        TransactionGen.TransactionStart(builder)

        if tid:
            TransactionGen.TransactionAddTid(builder, tid)

        if self.created:
            TransactionGen.TransactionAddCreated(builder, int(self.created))

        if self.created_payment_channel_seq:
            TransactionGen.TransactionAddCreatedPaymentChannelSeq(builder, self.created_payment_channel_seq)

        if self.created_paying_channel_seq:
            TransactionGen.TransactionAddCreatedPayingChannelSeq(builder, self.created_paying_channel_seq)

        if offer:
            TransactionGen.TransactionAddOffer(builder, offer)

        if amount:
            TransactionGen.TransactionAddAmount(builder, amount)

        if payment_channel:
            TransactionGen.TransactionAddPaymentChannel(builder, payment_channel)

        if paying_channel:
            TransactionGen.TransactionAddPayingChannel(builder, paying_channel)

        if self.state:
            TransactionGen.TransactionAddState(builder, int(self.state))

        if self.completed:
            TransactionGen.TransactionAddCompleted(builder, int(self.completed))

        if self.completed_payment_channel_seq:
            TransactionGen.TransactionAddCompletedPaymentChannelSeq(builder, self.completed_payment_channel_seq)

        if self.completed_paying_channel_seq:
            TransactionGen.TransactionAddCompletedPayingChannelSeq(builder, self.completed_paying_channel_seq)

        if key:
            TransactionGen.TransactionAddKey(builder, key)

        if buyer_pubkey:
            TransactionGen.TransactionAddBuyerPubkey(builder, buyer_pubkey)

        if payment_channel_after:
            TransactionGen.TransactionAddPaymentChannelAfter(builder, payment_channel_after)

        if paying_channel_after:
            TransactionGen.TransactionAddPayingChannelAfter(builder, paying_channel_after)

        if payment_mm_sig:
            TransactionGen.TransactionAddPaymentMmSig(builder, payment_mm_sig)

        if payment_del_sig:
            TransactionGen.TransactionAddPaymentDelSig(builder, payment_del_sig)

        if paying_mm_sig:
            TransactionGen.TransactionAddPayingMmSig(builder, paying_mm_sig)

        if paying_del_sig:
            TransactionGen.TransactionAddPayingDelSig(builder, paying_del_sig)

        final = TransactionGen.TransactionEnd(builder)

        return final


@table("87cf8eac-10d8-470d-a645-23e11343e065", build=Transaction.build, cast=Transaction.cast)
class Transactions(MapUuidFlatBuffers):
    """
    Data encryption key (off-chain) market transactions.
    """



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from datetime import datetime
from pprint import pformat
from typing import List, Optional
from uuid import UUID

from cfxdb.common import ConfigurationElement
from cfxdb.gen.user.OrganizationType import OrganizationType


class Organization(ConfigurationElement):
    """
    Organizations created in this master instance.
    """

    OTYPES = [OrganizationType.NONE, OrganizationType.BUSINESS, OrganizationType.ACADEMICS, OrganizationType.PERSONAL]
    """
    Organization type.
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        name: Optional[str] = None,
        otype: Optional[int] = None,
        registered: Optional[datetime] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of the organization

        :param label: Optional user label of the organization

        :param description: Optional user description of the organization

        :param tags: Optional list of user tags on the organization

        :param name: Name of the organization

        :param otype: Type of the organization.

        :param registered: Timestamp when the organization was created
        """
        ConfigurationElement.__init__(self, oid=oid, label=label, description=description, tags=tags)

        self.name = name
        self.otype = otype
        self.registered = registered

        # private member with unknown/untouched data passing through
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not ConfigurationElement.__eq__(self, other):
            return False
        if other.name != self.name:
            return False
        if other.otype != self.otype:
            return False
        if other.registered != self.registered:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "\n{}\n".format(pformat(self.marshal()))

    def copy(self, other):
        ConfigurationElement.copy(self, other)

        self.name = other.name
        self.otype = other.otype
        self.registered = other.registered

        # _unknown is not copied!

    def marshal(self):
        obj = ConfigurationElement.marshal(self)

        assert self.name is None or type(self.name) == str
        assert self.otype is None or self.otype in Organization.OTYPES
        assert self.registered is None or isinstance(self.registered, datetime)

        registered = int(self.registered.timestamp() * 1000000) if self.registered else None
        obj.update(
            {
                "name": self.name,
                "otype": self.otype,
                "registered": registered,
            }
        )

        if self._unknown:
            # pass through all attributes unknown
            obj.update(self._unknown)

        return obj

    @staticmethod
    def parse(data):
        assert type(data) == dict

        obj = ConfigurationElement.parse(data)
        data = obj._unknown

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["name", "otype", "registered"]:
                _unknown[k] = data[k]

        name = data.get("name", None)
        assert name is None or type(name) == str

        otype = data.get("otype", None)
        assert otype is None or otype in Organization.OTYPES

        registered = data.get("registered", None)
        assert registered is None or type(registered) == float or type(registered) == int
        if registered:
            # registered = datetime.utcfromtimestamp(float(registered) / 1000000.)
            registered = datetime.fromtimestamp(float(registered) / 1000000.0)

        obj = Organization(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            name=name,
            otype=otype,
            registered=registered,
            _unknown=_unknown,
        )
        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid

import numpy as np
from zlmdb import MapBytes32FlatBuffers, MapUuidTimestampBytes32, flatbuffers, table

from cfxdb.gen.xbrnetwork import UserKey as UserKeyGen


class _UserKeyGen(UserKeyGen.UserKey):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsUserKey(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _UserKeyGen()
        x.Init(buf, n + offset)
        return x

    def PubkeyAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def OwnerAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class UserKey(object):
    """
    User client (public) keys.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # User key - a Ed25519 public key - for authenticating using WAMP-cryptosign.
        # [uint8] (uint256)
        self._pubkey = None

        # Timestamp (epoch time in ns) of initial creation of this record.
        # uint64 (timestamp)
        self._created = None

        # ID of user account this user key is owned by.
        # [uint8] (uuid)
        self._owner = None

    def marshal(self):
        obj = {
            "pubkey": bytes(self.pubkey),
            "created": int(self.created),
            "owner": self.owner.bytes,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def pubkey(self) -> bytes:
        """
        User key - a Ed25519 public key - for authenticating using WAMP-cryptosign.
        """
        if self._pubkey is None and self._from_fbs:
            if self._from_fbs.PubkeyLength():
                self._pubkey = self._from_fbs.PubkeyAsBytes()
        return self._pubkey

    @pubkey.setter
    def pubkey(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 32)
        self._pubkey = value

    @property
    def created(self) -> np.datetime64:
        """
        Timestamp (epoch time in ns) of initial creation of this record.
        """
        if self._created is None and self._from_fbs:
            self._created = np.datetime64(self._from_fbs.Created(), "ns")
        return self._created

    @created.setter
    def created(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._created = value

    @property
    def owner(self) -> uuid.UUID:
        """
        ID of user account this user key is owned by.
        """
        if self._owner is None and self._from_fbs:
            if self._from_fbs.OwnerLength():
                _owner = self._from_fbs.OwnerAsBytes()
                self._owner = uuid.UUID(bytes=bytes(_owner))
            else:
                self._owner = uuid.UUID(bytes=b"\x00" * 20)
        return self._owner

    @owner.setter
    def owner(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._owner = value

    @staticmethod
    def cast(buf):
        return UserKey(_UserKeyGen.GetRootAsUserKey(buf, 0))

    def build(self, builder):
        owner = self.owner.bytes if self.owner else None
        if owner:
            owner = builder.CreateString(owner)

        pubkey = self.pubkey
        if pubkey:
            pubkey = builder.CreateString(pubkey)

        UserKeyGen.UserKeyStart(builder)

        if pubkey:
            UserKeyGen.UserKeyAddPubkey(builder, pubkey)

        if self.created:
            UserKeyGen.UserKeyAddCreated(builder, int(self.created))

        if owner:
            UserKeyGen.UserKeyAddOwner(builder, owner)

        final = UserKeyGen.UserKeyEnd(builder)

        return final


@table("5b5d0ce7-33f4-4421-a6ab-ed77cafc763a", build=UserKey.build, cast=UserKey.cast)
class UserKeys(MapBytes32FlatBuffers):
    """
    Database table for user client keys.
    """

    @staticmethod
    def parse(data):
        """

        :param data:
        :return:
        """
        pubkey = None
        if "pubkey" in data:
            assert type(data["pubkey"]) == bytes and len(data["pubkey"]) == 32
            pubkey = data["pubkey"]

        created = None
        if "created" in data:
            assert type(data["created"] == int)
            created = np.datetime64(data["created"], "ns")

        owner = None
        if "owner" in data:
            assert type(data["owner"] == bytes and len(data["owner"]) == 16)
            owner = uuid.UUID(bytes=data["owner"])

        obj = UserKey()
        obj.pubkey = pubkey
        obj.created = created
        obj.owner = owner

        return obj


@table("68b736f8-27df-4e3e-b80f-1b855ae5596f")
class IndexUserKeyByAccount(MapUuidTimestampBytes32):
    """
    Database (index) table for (member_oid, created) -> userkey mapping.
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrmm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Data encryption key (off-chain) market transactions.
class Transaction(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Transaction()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTransaction(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Transaction
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ID of the transaction.
    # Transaction
    def Tid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def TidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def TidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def TidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Creation time of the transaction (epoch time in ns).
    # Transaction
    def Created(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Sequence number of the created-state transaction within the payment channel.
    # Transaction
    def CreatedPaymentChannelSeq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Sequence number of the created-state transaction within the paying channel.
    # Transaction
    def CreatedPayingChannelSeq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # ID of the data encryption key offer this transaction is for.
    # Transaction
    def Offer(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def OfferAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def OfferLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def OfferIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Transaction amount in ERC20 tokens of the market coin type.
    # Transaction
    def Amount(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def AmountAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def AmountLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def AmountIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # OID of the channel (of the buyer) this transaction is transacting on.
    # Transaction
    def PaymentChannel(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def PaymentChannelAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def PaymentChannelLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def PaymentChannelIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # OID of the channel (of the seller) this transaction is transacting on.
    # Transaction
    def PayingChannel(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def PayingChannelAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def PayingChannelLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def PayingChannelIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # State of the transaction: TransactionState.INFLIGHT when created, and then either TransactionState.SUCCESS or TransactionState.FAILED.
    # Transaction
    def State(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Completion time of the transaction (epoch time in ns)
    # Transaction
    def Completed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Sequence number of the completed-state transaction within the payment channel.
    # Transaction
    def CompletedPaymentChannelSeq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Sequence number of the completed-state transaction within the paying channel.
    # Transaction
    def CompletedPayingChannelSeq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # ID of the data encryption key sold under the transaction.
    # Transaction
    def Key(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def KeyAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def KeyLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def KeyIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        return o == 0

    # Buyer public key (Ed25519).
    # Transaction
    def BuyerPubkey(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def BuyerPubkeyAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def BuyerPubkeyLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def BuyerPubkeyIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        return o == 0

    # Payment channel balance after transaction.
    # Transaction
    def PaymentChannelAfter(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def PaymentChannelAfterAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def PaymentChannelAfterLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def PaymentChannelAfterIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        return o == 0

    # Payment channel balance after transaction.
    # Transaction
    def PayingChannelAfter(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def PayingChannelAfterAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def PayingChannelAfterLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def PayingChannelAfterIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        return o == 0

    # Payment channel market maker transaction signature.
    # Transaction
    def PaymentMmSig(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def PaymentMmSigAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def PaymentMmSigLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def PaymentMmSigIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        return o == 0

    # Payment channel (buyer) delegate transaction signature.
    # Transaction
    def PaymentDelSig(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def PaymentDelSigAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def PaymentDelSigLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def PaymentDelSigIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        return o == 0

    # Paying channel market maker transaction signature.
    # Transaction
    def PayingMmSig(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def PayingMmSigAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def PayingMmSigLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def PayingMmSigIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        return o == 0

    # Paying channel (seller) delegate transaction signature.
    # Transaction
    def PayingDelSig(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Transaction
    def PayingDelSigAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Transaction
    def PayingDelSigLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Transaction
    def PayingDelSigIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        return o == 0


def TransactionStart(builder):
    builder.StartObject(20)


def Start(builder):
    TransactionStart(builder)


def TransactionAddTid(builder, tid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(tid), 0)


def AddTid(builder, tid):
    TransactionAddTid(builder, tid)


def TransactionStartTidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTidVector(builder, numElems: int) -> int:
    return TransactionStartTidVector(builder, numElems)


def TransactionAddCreated(builder, created):
    builder.PrependUint64Slot(1, created, 0)


def AddCreated(builder, created):
    TransactionAddCreated(builder, created)


def TransactionAddCreatedPaymentChannelSeq(builder, createdPaymentChannelSeq):
    builder.PrependUint32Slot(2, createdPaymentChannelSeq, 0)


def AddCreatedPaymentChannelSeq(builder, createdPaymentChannelSeq):
    TransactionAddCreatedPaymentChannelSeq(builder, createdPaymentChannelSeq)


def TransactionAddCreatedPayingChannelSeq(builder, createdPayingChannelSeq):
    builder.PrependUint32Slot(3, createdPayingChannelSeq, 0)


def AddCreatedPayingChannelSeq(builder, createdPayingChannelSeq):
    TransactionAddCreatedPayingChannelSeq(builder, createdPayingChannelSeq)


def TransactionAddOffer(builder, offer):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(offer), 0)


def AddOffer(builder, offer):
    TransactionAddOffer(builder, offer)


def TransactionStartOfferVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOfferVector(builder, numElems: int) -> int:
    return TransactionStartOfferVector(builder, numElems)


def TransactionAddAmount(builder, amount):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(amount), 0)


def AddAmount(builder, amount):
    TransactionAddAmount(builder, amount)


def TransactionStartAmountVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAmountVector(builder, numElems: int) -> int:
    return TransactionStartAmountVector(builder, numElems)


def TransactionAddPaymentChannel(builder, paymentChannel):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(paymentChannel), 0)


def AddPaymentChannel(builder, paymentChannel):
    TransactionAddPaymentChannel(builder, paymentChannel)


def TransactionStartPaymentChannelVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPaymentChannelVector(builder, numElems: int) -> int:
    return TransactionStartPaymentChannelVector(builder, numElems)


def TransactionAddPayingChannel(builder, payingChannel):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(payingChannel), 0)


def AddPayingChannel(builder, payingChannel):
    TransactionAddPayingChannel(builder, payingChannel)


def TransactionStartPayingChannelVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPayingChannelVector(builder, numElems: int) -> int:
    return TransactionStartPayingChannelVector(builder, numElems)


def TransactionAddState(builder, state):
    builder.PrependUint8Slot(8, state, 0)


def AddState(builder, state):
    TransactionAddState(builder, state)


def TransactionAddCompleted(builder, completed):
    builder.PrependUint64Slot(9, completed, 0)


def AddCompleted(builder, completed):
    TransactionAddCompleted(builder, completed)


def TransactionAddCompletedPaymentChannelSeq(builder, completedPaymentChannelSeq):
    builder.PrependUint32Slot(10, completedPaymentChannelSeq, 0)


def AddCompletedPaymentChannelSeq(builder, completedPaymentChannelSeq):
    TransactionAddCompletedPaymentChannelSeq(builder, completedPaymentChannelSeq)


def TransactionAddCompletedPayingChannelSeq(builder, completedPayingChannelSeq):
    builder.PrependUint32Slot(11, completedPayingChannelSeq, 0)


def AddCompletedPayingChannelSeq(builder, completedPayingChannelSeq):
    TransactionAddCompletedPayingChannelSeq(builder, completedPayingChannelSeq)


def TransactionAddKey(builder, key):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(key), 0)


def AddKey(builder, key):
    TransactionAddKey(builder, key)


def TransactionStartKeyVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartKeyVector(builder, numElems: int) -> int:
    return TransactionStartKeyVector(builder, numElems)


def TransactionAddBuyerPubkey(builder, buyerPubkey):
    builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(buyerPubkey), 0)


def AddBuyerPubkey(builder, buyerPubkey):
    TransactionAddBuyerPubkey(builder, buyerPubkey)


def TransactionStartBuyerPubkeyVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartBuyerPubkeyVector(builder, numElems: int) -> int:
    return TransactionStartBuyerPubkeyVector(builder, numElems)


def TransactionAddPaymentChannelAfter(builder, paymentChannelAfter):
    builder.PrependUOffsetTRelativeSlot(14, flatbuffers.number_types.UOffsetTFlags.py_type(paymentChannelAfter), 0)


def AddPaymentChannelAfter(builder, paymentChannelAfter):
    TransactionAddPaymentChannelAfter(builder, paymentChannelAfter)


def TransactionStartPaymentChannelAfterVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPaymentChannelAfterVector(builder, numElems: int) -> int:
    return TransactionStartPaymentChannelAfterVector(builder, numElems)


def TransactionAddPayingChannelAfter(builder, payingChannelAfter):
    builder.PrependUOffsetTRelativeSlot(15, flatbuffers.number_types.UOffsetTFlags.py_type(payingChannelAfter), 0)


def AddPayingChannelAfter(builder, payingChannelAfter):
    TransactionAddPayingChannelAfter(builder, payingChannelAfter)


def TransactionStartPayingChannelAfterVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPayingChannelAfterVector(builder, numElems: int) -> int:
    return TransactionStartPayingChannelAfterVector(builder, numElems)


def TransactionAddPaymentMmSig(builder, paymentMmSig):
    builder.PrependUOffsetTRelativeSlot(16, flatbuffers.number_types.UOffsetTFlags.py_type(paymentMmSig), 0)


def AddPaymentMmSig(builder, paymentMmSig):
    TransactionAddPaymentMmSig(builder, paymentMmSig)


def TransactionStartPaymentMmSigVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPaymentMmSigVector(builder, numElems: int) -> int:
    return TransactionStartPaymentMmSigVector(builder, numElems)


def TransactionAddPaymentDelSig(builder, paymentDelSig):
    builder.PrependUOffsetTRelativeSlot(17, flatbuffers.number_types.UOffsetTFlags.py_type(paymentDelSig), 0)


def AddPaymentDelSig(builder, paymentDelSig):
    TransactionAddPaymentDelSig(builder, paymentDelSig)


def TransactionStartPaymentDelSigVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPaymentDelSigVector(builder, numElems: int) -> int:
    return TransactionStartPaymentDelSigVector(builder, numElems)


def TransactionAddPayingMmSig(builder, payingMmSig):
    builder.PrependUOffsetTRelativeSlot(18, flatbuffers.number_types.UOffsetTFlags.py_type(payingMmSig), 0)


def AddPayingMmSig(builder, payingMmSig):
    TransactionAddPayingMmSig(builder, payingMmSig)


def TransactionStartPayingMmSigVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPayingMmSigVector(builder, numElems: int) -> int:
    return TransactionStartPayingMmSigVector(builder, numElems)


def TransactionAddPayingDelSig(builder, payingDelSig):
    builder.PrependUOffsetTRelativeSlot(19, flatbuffers.number_types.UOffsetTFlags.py_type(payingDelSig), 0)


def AddPayingDelSig(builder, payingDelSig):
    TransactionAddPayingDelSig(builder, payingDelSig)


def TransactionStartPayingDelSigVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPayingDelSigVector(builder, numElems: int) -> int:
    return TransactionStartPayingDelSigVector(builder, numElems)


def TransactionEnd(builder):
    return builder.EndObject()


def End(builder):
    return TransactionEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import binascii
import pprint
from typing import Optional
from uuid import UUID

import numpy as np

from cfxdb._exception import InvalidConfigException


class Credential(object):
    """
    Credentials created for use with WAMP authentication.
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        created: Optional[np.datetime64] = None,
        authmethod: Optional[str] = None,
        authid: Optional[str] = None,
        realm: Optional[str] = None,
        authconfig: Optional[dict] = None,
        principal_oid: Optional[UUID] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of this credential object

        :param created: Timestamp when credential was created.

        :param authmethod: WAMP authentication method offered by the authenticating client.

        :param realm: WAMP realm requested by the authenticating client.

        :param authid: WAMP authid announced by the authenticating client.

        :param authconfig: Authentication method specific configuration.

        :param principal_oid: ID of the principal this credential resolves to upon successful authentication.
        """
        self.oid = oid
        self.created = created
        self.authmethod = authmethod
        self.realm = realm
        self.authid = authid
        self.authconfig = authconfig
        self.principal_oid = principal_oid

        # private member with unknown/untouched data passing through
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if other.oid != self.oid:
            return False
        if other.created != self.created:
            return False
        if other.authmethod != self.authmethod:
            return False
        if other.realm != self.realm:
            return False
        if other.authid != self.authid:
            return False
        if other.authconfig != self.authconfig:
            return False
        if other.principal_oid != self.principal_oid:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def copy(self, other, overwrite=False):
        """
        Copy over other object.

        :param other: Other credential to copy data from.
        :type other: instance of :class:`ManagementRealm`
        :return:
        """
        if (not self.oid and other.oid) or overwrite:
            self.oid = other.oid
        if (not self.created and other.created) or overwrite:
            self.created = other.created
        if (not self.authmethod and other.authmethod) or overwrite:
            self.authmethod = other.authmethod
        if (not self.realm and other.realm) or overwrite:
            self.realm = other.realm
        if (not self.authid and other.authid) or overwrite:
            self.authid = other.authid
        if (not self.authconfig and other.authconfig) or overwrite:
            self.authconfig = other.authconfig
        if (not self.principal_oid and other.principal_oid) or overwrite:
            self.principal_oid = other.principal_oid

        # _unknown is not copied!

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = {
            "oid": str(self.oid) if self.oid else None,
            "created": int(self.created) if self.created else None,
            "authmethod": self.authmethod,
            "realm": self.realm,
            "authid": self.authid,
            "authconfig": self.authconfig,
            "principal_oid": str(self.principal_oid) if self.principal_oid else None,
        }

        if self._unknown:
            # pass through all attributes unknown
            obj.update(self._unknown)

        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`ManagementRealm`
        """
        assert type(data) == dict

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["oid", "created", "authmethod", "realm", "authid", "authconfig", "principal_oid"]:
                _unknown[k] = data[k]

        oid = data.get("oid", None)
        assert oid is None or type(oid) == str
        if oid:
            oid = UUID(oid)

        created = data.get("created", None)
        assert created is None or type(created) == int
        if created:
            created = np.datetime64(created, "ns")

        authmethod = data.get("authmethod", None)
        assert authmethod is None or type(authmethod) == str

        realm = data.get("realm", None)
        assert realm is None or type(realm) == str

        authid = data.get("authid", None)
        assert authid is None or type(authid) == str

        principal_oid = data.get("principal_oid", None)
        assert principal_oid is None or type(principal_oid) == str
        if principal_oid:
            principal_oid = UUID(principal_oid)

        authconfig = data.get("authconfig", None)
        assert authconfig is None or type(authconfig) == dict

        if authconfig:
            if authmethod == "cryptosign":
                # check allowed keys
                for key in authconfig:
                    if key not in ["authorized_keys"]:
                        raise InvalidConfigException('invalid attribute "{}" in cryptosign config'.format(key))

                # check required keys
                if "authorized_keys" not in authconfig or type(authconfig["authorized_keys"]) != list:
                    raise InvalidConfigException(
                        'invalid type "{}" for authorized_keys in cryptosign config'.format(
                            type(authconfig["authorized_keys"])
                        )
                    )
                if len(authconfig["authorized_keys"]) == 0:
                    raise InvalidConfigException("need at least one key in authorized_keys in cryptosign config")
                authorized_keys = authconfig["authorized_keys"]
                for k in authorized_keys:
                    if type(k) != str or len(k) != 64:
                        raise InvalidConfigException(
                            "key in autheorized_keys must have type str[64] (was {}) in authorized_keys in cryptosign config".format(
                                type(k)
                            )
                        )
                    try:
                        binascii.a2b_hex(k)
                    except Exception as e:
                        raise InvalidConfigException(
                            "invalid key in autheorized_keys in authorized_keys in cryptosign config: {}".format(e)
                        )

            elif authmethod == "ticket":
                # check allowed keys
                for key in authconfig:
                    if key not in ["secret"]:
                        raise InvalidConfigException('invalid attribute "{}" in ticket config'.format(key))

                # check required keys
                if "secret" not in authconfig or type(authconfig["secret"]) != str:
                    raise InvalidConfigException(
                        'invalid type "{}" for authorized_keys in ticket config'.format(
                            type(authconfig["authorized_keys"])
                        )
                    )

            elif authmethod == "wampcra":
                # check allowed keys
                for key in authconfig:
                    if key not in ["secret", "salt", "iterations", "keylen"]:
                        raise InvalidConfigException('invalid attribute "{}" in wampcra config'.format(key))

                # check required keys
                if "secret" not in authconfig or type(authconfig["secret"]) != str:
                    raise InvalidConfigException(
                        'invalid type "{}" for secret in wampcra config'.format(type(authconfig["secret"]))
                    )
                if "salt" in authconfig and type(authconfig["salt"]) != str:
                    raise InvalidConfigException(
                        'invalid type "{}" for salt in wampcra config'.format(type(authconfig["salt"]))
                    )
                if "iterations" in authconfig and type(authconfig["iterations"]) != int:
                    raise InvalidConfigException(
                        'invalid type "{}" for iterations in wampcra config'.format(type(authconfig["iterations"]))
                    )
                if "keylen" in authconfig and type(authconfig["keylen"]) != int:
                    raise InvalidConfigException(
                        'invalid type "{}" for keylen in wampcra config'.format(type(authconfig["keylen"]))
                    )

            elif authmethod == "tls":
                raise NotImplementedError("FIXME: check tls authmethod configuration")

            elif authmethod == "scram":
                # check allowed keys
                for key in authconfig:
                    if key not in ["kdf", "iterations", "memory", "salt", "stored-key", "server-key"]:
                        raise InvalidConfigException('invalid attribute "{}" in scram config'.format(key))

                # check required keys
                for key, Type in [
                    ("kdf", str),
                    ("iterations", int),
                    ("memory", int),
                    ("salt", str),
                    ("stored-key", str),
                    ("server-key", str),
                ]:
                    if key not in authconfig or type(authconfig[key]) != Type:
                        raise InvalidConfigException(
                            'invalid type "{}" for secret in scram config'.format(type(authconfig[key]))
                        )

            elif authmethod == "cookie":
                raise NotImplementedError("FIXME: check cookie authmethod configuration")

            elif authmethod == "anonymous":
                # there is nothing to configure for authmethod==anonymous (authid/authrole actually assigned
                # is defined already "outside" this config dict)
                for key in authconfig:
                    if key not in []:
                        raise InvalidConfigException('invalid attribute "{}" in anonymous config'.format(key))
            else:
                raise InvalidConfigException('invalid authmethod "{}"'.format(authmethod))

        obj = Credential(
            oid=oid,
            created=created,
            authmethod=authmethod,
            realm=realm,
            authid=authid,
            authconfig=authconfig,
            principal_oid=principal_oid,
            _unknown=_unknown,
        )

        return obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log

# CFC instance (master node) run-time (live) states.
class MRealmState(object):
    # Unset state.
    NONE = 0
    # Online and fully operational.
    ONLINE = 1
    # Online, but some error occurred and some features or entities might be unavailable.
    DEGRADED = 2
    # Ended gracefully.
    ENDED = 3
    # The state was determined as "not current" (eg when the previous running CFC was hard-killed) and the record was cleaned up (marked as ended).
    PURGED = 4



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import shutil
import tempfile
import uuid

import numpy as np
import pytest
import txaio
from txaio import time_ns

txaio.use_twisted()

import zlmdb

from cfxdb.globalschema import GlobalSchema
from cfxdb.usage import MasterNodeUsage


def fill_usage(usage):
    usage.timestamp = np.datetime64(time_ns(), "ns")
    usage.mrealm_id = uuid.uuid4()
    usage.timestamp_from = np.datetime64(usage.timestamp - np.timedelta64(10, "m"), "ns")
    usage.pubkey = os.urandom(32)

    usage.client_ip_version = random.choice([4, 6])
    if usage.client_ip_version == 4:
        usage.client_ip_address = os.urandom(4)
    else:
        usage.client_ip_address = os.urandom(16)
    usage.client_ip_port = random.randint(1, 2**16 - 1)

    usage.seq = random.randint(0, 1000000)
    usage.sent = np.datetime64(time_ns() - random.randint(0, 10**10), "ns")
    usage.processed = np.datetime64(time_ns() + random.randint(0, 10**10), "ns")
    usage.status = random.randint(0, 3)
    usage.status_message = "hello world {}".format(uuid.uuid4())
    usage.metering_id = uuid.uuid4()

    usage.count = random.randint(0, 100000)
    usage.total = random.randint(0, 100000)
    usage.nodes = random.randint(0, 100000)

    usage.controllers = random.randint(0, 100000)
    usage.hostmonitors = random.randint(0, 100000)
    usage.routers = random.randint(0, 100000)
    usage.containers = random.randint(0, 100000)
    usage.guests = random.randint(0, 100000)
    usage.proxies = random.randint(0, 100000)
    usage.marketmakers = random.randint(0, 100000)

    usage.sessions = random.randint(0, 100000)

    usage.msgs_call = random.randint(0, 100000)
    usage.msgs_yield = random.randint(0, 100000)
    usage.msgs_invocation = random.randint(0, 100000)
    usage.msgs_result = random.randint(0, 100000)
    usage.msgs_error = random.randint(0, 100000)
    usage.msgs_publish = random.randint(0, 100000)
    usage.msgs_published = random.randint(0, 100000)
    usage.msgs_event = random.randint(0, 100000)
    usage.msgs_register = random.randint(0, 100000)
    usage.msgs_registered = random.randint(0, 100000)
    usage.msgs_subscribe = random.randint(0, 100000)
    usage.msgs_subscribed = random.randint(0, 100000)


@pytest.fixture(scope="module")
def db(scratch=True):
    dbpath = os.path.join(tempfile.gettempdir(), "testdb")
    if scratch and os.path.exists(dbpath):
        shutil.rmtree(dbpath)
    db = zlmdb.Database(dbpath=dbpath, writemap=True)
    db.__enter__()
    return db


def test_stats(db):
    _dbs = GlobalSchema.attach(db)  # noqa: F841

    # {'branch_pages': 0,
    # 'current_size': 10485760,
    # 'depth': 1,
    # 'entries': 14,
    # 'free': 0.999609375,
    # 'last_pgno': 7,
    # 'last_txnid': 14,
    # 'leaf_pages': 1,
    # 'map_addr': 0,
    # 'map_size': 10485760,
    # 'max_readers': 126,
    # 'max_size': 10485760,
    # 'num_readers': 1,
    # 'overflow_pages': 0,
    # 'pages': 1,
    # 'pages_size': 4096,
    # 'psize': 4096,
    # 'read_only': False,
    # 'sync_enabled': True,
    # 'zlmdb_slots': 14}
    stats = db.stats()

    # check default maximum size
    assert stats["max_size"] == 10485760

    # check current size, which is maxsize when writemap==True (which it is by default)
    assert stats["current_size"] == 10485760

    # however, the DB is empty ..
    assert stats["pages"] == 1
    assert stats["free"] == 0.999609375

    # GlobalSchema has 14 tables
    assert stats["num_slots"] == 14


def test_usage_stats(db):
    dbs: GlobalSchema = GlobalSchema.attach(db)

    stats_begin = db.stats()
    assert stats_begin["pages"] == 1
    assert stats_begin["free"] == 0.999609375

    with db.begin(write=True) as txn:
        for i in range(10000):
            usage = MasterNodeUsage()
            fill_usage(usage)
            dbs.usage[txn, (usage.timestamp, usage.mrealm_id)] = usage

    stats_end = db.stats()
    assert stats_end["pages"] == 1684
    assert stats_end["free"] == 0.3421875



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: user

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# CFC user roles on a management realm
class UserMrealmRoles(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = UserMrealmRoles()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsUserMrealmRoles(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # UserMrealmRoles
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # List of roles the user has on the respective mrealm.
    # UserMrealmRoles
    def Roles(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Int8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # UserMrealmRoles
    def RolesAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int8Flags, o)
        return 0

    # UserMrealmRoles
    def RolesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # UserMrealmRoles
    def RolesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0


def UserMrealmRolesStart(builder):
    builder.StartObject(1)


def Start(builder):
    UserMrealmRolesStart(builder)


def UserMrealmRolesAddRoles(builder, roles):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(roles), 0)


def AddRoles(builder, roles):
    UserMrealmRolesAddRoles(builder, roles)


def UserMrealmRolesStartRolesVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartRolesVector(builder, numElems: int) -> int:
    return UserMrealmRolesStartRolesVector(builder, numElems)


def UserMrealmRolesEnd(builder):
    return builder.EndObject()


def End(builder):
    return UserMrealmRolesEnd(builder)
